part of selene;

/// A storage class to hold the current state of a [Client], and it's cached objects.
class ClientState {
  /// The [Client] that this state represents.
  Client client;

  /// The current connected Discord user.
  ///
  /// This will be `null` if the user hasn't authenticated yet.
  User currentUser;

  /// All users that this client has encountered or received.
  ///
  /// The key is an [int], representing their snowflake ID.
  ///
  /// This contains the current connected user.
  Map<int, User> users = {};

  /// All guild members that this client has encountered or received.
  ///
  /// The key is an [int], representing the ID of the user object that it wraps.
  ///
  /// This contains the current connected user (may contain multiple if the
  /// user is in multiple guilds).
  Map<int, GuildMember> members = {};

  /// All guilds that this client is a member of.
  ///
  /// The key is an [int], representing the ID of the guild.
  Map<int, Guild> guilds = {};

  /// All guild channels that this client has encountered or received.
  ///
  /// The key is an [int], representing the snowflake ID of the channel.
  Map<int, GuildChannel> guildChannels = {};

  /// All two-way DM/private channels that this client has received/sent.
  ///
  /// The key is an [int], representing the ID of the channel.
  Map<int, DMChannel> dmChannels = {};

  /// All guild emotes that this client has seen as part of guild objects.
  ///
  /// The key is an [int], representing the emote's snowflake ID.
  Map<int, GuildEmote> guildEmotes = {};

  /// All guild roles that this client has encountered or seen.
  ///
  /// The key is an [int], representing the guild's ID.
  Map<int, GuildRole> roles = {};

  /// All voice regions that have been downloaded.
  ///
  /// The key is a [String], representing the voice region's descriptor/ID.
  Map<String, VoiceRegion> voiceRegions = {};

  /// Creates an empty [ClientState].
  ClientState(this.client);

  /// Resolves a [GuildChannel] or [DMChannel] object from this state.
  ///
  /// Returns null if the cache record does not exist.
  Channel resolveChannel(int id) {
    if (guildChannels.containsKey(id))
      return guildChannels[id];
    else if (dmChannels.containsKey(id))
      return dmChannels[id];
    else
      return null; // Invalid key or unknown channel
  }
}
