part of selene;

/// A Discord client.
class Client {
  /// The current event dispatcher.
  Dispatcher dispatcher;

  /// The current REST API implementation.
  RestApiBase _rest;

  /// The current WebSocket API implementation.
  WSBase _ws;

  /// The current JSON client.
  json.JsonClient _json;

  /// The current cache for all Discord entities.
  ClientState state;

  /// Shortcut for `state.guildChannels`
  List<GuildChannel> get guildChannels => state.guildChannels.values;

  /// Shortcut for `state.dmChannels`
  List<DMChannel> get dmChannels => state.dmChannels.values;

  /// Combines `guildChannels` and `dmChannels` into one list of all channels this client is connected to.
  List<Channel> get channels =>
      new List.from(
          guildChannels.cast<Channel>().followedBy(dmChannels.cast<Channel>()));

  /// Shortcut for `state.guilds`
  List<Guild> get guilds => state.guilds.values;

  /// Shortcut for `state.members`
  List<GuildMember> get members => state.members.values;

  /// Shortcut for `state.users`
  List<User> get users => state.users.values;

  /// Shortcut for `state.voiceRegions`
  List<VoiceRegion> get voiceRegions => state.voiceRegions.values;

  /// The current [ClientOptions] for this client.
  ClientOptions options;

  /// Initializes this client.
  Client({ClientOptions options}) {
    this.options = options ?? new ClientOptions();
    _rest = this.options.rest ?? new _RestApiImpl();
    _ws = this.options.ws ?? new _WebSocketImpl(this);
    dispatcher = new Dispatcher(this);
    state = new ClientState(this);
    _json = new json.JsonClient();
  }

  /// The current logged-in client user.
  ///
  /// Will be `null` if the client hasn't authenticated.
  User get currentUser => state.currentUser;

  /// Authenticates this Discord client using the provided token and type.
  ///
  /// Must be invoked before [Client.start] has been invoked to ensure a successful connection.
  void login(String token, {String type = 'Bot'}) {
    if (token == null || token.isEmpty)
      throw new ArgumentError('Token provided must be not null or empty');
    if (type == null || !['Bot', '', 'Bearer', 'Webhook'].contains(type))
      throw new ArgumentError('Provided token type is invalid or null');

    _rest.login(token, type: type);
    _ws.login(token, type: type);
  }

  /// Opens the WebSocket connection.
  ///
  /// Before this, [Client.login] should be invoked to store authorization data.
  ///
  /// `voiceRegionDownloadOptions` can be specified to customise the voice
  /// region data download.
  Future start([RequestOptions voiceRegionDownloadOptions]) async {
    await _ws.start(false);
    if (options.downloadVoiceRegions) state.voiceRegions =
    await downloadVoiceRegions(voiceRegionDownloadOptions);
  }

  /// Closes the WebSocket connection.
  ///
  /// Calling anything on the WebSocket will now throw an exception.
  Future stop() => _ws.stop();

  // REST endpoints

  /// Downloads voice region data.
  ///
  /// This is automatically called when [Client.start] is invoked (if `options.downloadVoiceRegions` is true).
  Future<Map<String, VoiceRegion>> downloadVoiceRegions(
      [RequestOptions options]) async {
    var response =
    await _rest.makeNewRequest('/voice/regions', 'GET', null, options);
    var voiceRegions = <String, VoiceRegion>{};
    for (var jsonVoiceRegion in response.body.asJson()) {
      var voiceRegion = new VoiceRegion._create(jsonVoiceRegion, this);
      voiceRegions[voiceRegion.id] = voiceRegion;
    }
    return voiceRegions;
  }

  /// Creates a new guild.
  ///
  /// All properties must be specified.
  Future<Guild> createGuild(String name,
      VoiceRegion region,
      String imageBase64_128x128_jpg,
      int verificationLevel,
      int defaultMessageNotifications,
      int explicitContentFilter,
      [RequestOptions options]) async {
    if (state.guilds.length > 10) {
      // API Limitation - https://discordapp.com/developers/docs/resources/guild#create-guild
      throw new UnsupportedError(
          'Client cannot create guilds when in more than 10 guilds');
    }

    var guildProps = <String, dynamic>{
      'name': name,
      'region': region.id,
      'icon': imageBase64_128x128_jpg,
      'verification_level': verificationLevel,
      'default_message_notifications': defaultMessageNotifications,
      'explicit_content_filter': explicitContentFilter
    };

    var response =
    await _rest.makeNewRequest('/guilds', 'POST', guildProps, options);

    return state.guilds[parseInt(response.body.asJson()['id'])];
  }
}

/// An information container for customizing [Client] objects.
class ClientOptions {
  /// The WebSocket implementation to use. Defaults to the internal WS implementation.
  WSBase ws;

  /// The REST implementation to use. Defaults to the internal REST implementation.
  RestApiBase rest;

  /// The maximum cache size for [TextChannel] message caches.
  int messageCacheMaxSize = 100;

  /// Whether to download voice region data on client startup.
  bool downloadVoiceRegions = true;
}
