part of selene;

typedef EventHandler = Future Function(Map<String, dynamic> data);

/// An event dispatcher, which consumes and emits Discord entity events.
class Dispatcher {
  /// A map of all event handlers.
  Map<String, EventHandler> _eventHandlers;

  /// The [Client] this dispatcher belongs to.
  Client client;

  // Misc Events

  /// Emitted when the client is ready.
  Stream get onReady => _onReadyController.stream;
  StreamController _onReadyController = new StreamController.broadcast();

  // Guilds

  /// Emitted when the client joins a guild.
  Stream<Guild> get onGuildJoin => _onGuildJoinController.stream;
  StreamController<Guild> _onGuildJoinController =
      new StreamController.broadcast();

  /// Emitted when a guild is updated.
  Stream<Guild> get onGuildUpdate => _onGuildUpdateController.stream;
  StreamController<Guild> _onGuildUpdateController =
      new StreamController.broadcast();

  /// Emitted when the client leaves, or is removed from, a guild.
  Stream<Guild> get onGuildLeave => _onGuildLeaveController.stream;
  StreamController<Guild> _onGuildLeaveController =
      new StreamController.broadcast();

  /// Emitted when a new user joins a guild.
  Stream<GuildMember> get onGuildMemberJoin =>
      _onGuildMemberJoinController.stream;
  StreamController<GuildMember> _onGuildMemberJoinController =
  new StreamController.broadcast();

  /// Emitted when a guild member's information is updated.
  Stream<GuildMember> get onGuildMemberUpdate =>
      _onGuildMemberUpdateController.stream;
  StreamController<GuildMember> _onGuildMemberUpdateController =
  new StreamController.broadcast();

  /// Emitted when a guild member leaves a guild.
  Stream<GuildMember> get onGuildMemberLeave =>
      _onGuildMemberLeaveController.stream;
  StreamController<GuildMember> _onGuildMemberLeaveController =
  new StreamController.broadcast();

  /// Emitted when a role is created in a guild.
  Stream<GuildRole> get onGuildRoleCreate =>
      _onGuildRoleCreateController.stream;
  StreamController<GuildRole> _onGuildRoleCreateController =
  new StreamController.broadcast();

  /// Emitted when a role is updated in a guild.
  Stream<GuildRole> get onGuildRoleUpdate =>
      _onGuildRoleUpdateController.stream;
  StreamController<GuildRole> _onGuildRoleUpdateController =
  new StreamController.broadcast();

  /// Emitted when a role is deleted from a guild.
  Stream<GuildRole> get onGuildRoleDelete =>
      _onGuildRoleDeleteController.stream;
  StreamController<GuildRole> _onGuildRoleDeleteController =
  new StreamController.broadcast();

  // Channels

  /// Emitted when a channel is created that is relevant to the current user.
  Stream<Channel> get onChannelCreate => _onChannelCreateController.stream;
  StreamController<Channel> _onChannelCreateController =
  new StreamController.broadcast();

  /// Emitted when a channel is updated.
  Stream<Channel> get onChannelUpdate => _onChannelUpdateController.stream;
  StreamController<Channel> _onChannelUpdateController =
  new StreamController.broadcast();

  /// Emitted when a channel is deleted.
  Stream<Channel> get onChannelDelete => _onChannelDeleteController.stream;
  StreamController<Channel> _onChannelDeleteController =
  new StreamController.broadcast();

  // Messages

  /// Emitted when a message is received.
  Stream<Message> get onMessageReceive => _onMessageReceiveController.stream;
  StreamController<Message> _onMessageReceiveController =
  new StreamController.broadcast();

  /// Emitted when a message is updated.
  Stream<Message> get onMessageUpdate => _onMessageUpdateController.stream;
  StreamController<Message> _onMessageUpdateController =
  new StreamController.broadcast();

  /// Emitted when a message is deleted.
  Stream get onMessageDelete => _onMessageDeleteController.stream;
  StreamController _onMessageDeleteController =
  new StreamController.broadcast();

  /// Emitted when a bulk array of messages have been deleted.
  Stream<List<Message>> get onMessageBulkDelete =>
      _onMessageBulkDeleteController.stream;
  StreamController<List<Message>> _onMessageBulkDeleteController =
      new StreamController.broadcast();

  /// Creates a new dispatcher, and attaches events.
  Dispatcher(this.client) {
    _eventHandlers = {
      'READY': (data) async {
        for (var guildJson in data['guilds'] ?? []) {
          new Guild._lazy(guildJson, client);
        }
        if (data['user'] != null) {
          client.state.currentUser =
          new User._cachedOrNew(data['user'], client);
        }
        for (var privateChannelJson in data['private_channels'] ?? []) {
          new DMChannel._create(privateChannelJson, client);
        }
        _onReadyController.add(null);
      },
      'GUILD_CREATE': (data) async {
        var cacheResult = new Guild._cached(data, client);
        if (cacheResult != null) {
          // Lazy loading guild from READY or a guild becomes available again to client
          cacheResult._onCreate(data);
          // Don't raise guild join because we aren't joining, only receiving back filled data
        } else {
          var guild = new Guild._create(data, client);
          _onGuildJoinController.add(guild);
        }
      },
      'GUILD_UPDATE': (data) async {
        var guild = new Guild._cached(data, client);
        guild._update(data);
        _onGuildUpdateController.add(guild);
      },
      'GUILD_DELETE': (data) async {
        var guild = client.state.guilds[int.tryParse(data['id'])];
        if (data['unavailable'] == null) {
          // Removed or left guild
          client.state.guilds.remove(int.tryParse(data['id']));
          _onGuildLeaveController.add(guild);
        } else {
          // Guild has gone offline/unavailable
          guild._update(data);
        }
      },
      'CHANNEL_CREATE': (data) async {
        var type = Channel._getType(data);
        Guild guild;
        if (type == ChannelType.GuildText || type == ChannelType.GuildVoice) {
          guild = new Guild._cached({'id': data['guild_id']}, client);
        }
        var channel = Channel._create(data, client, guild);
        _onChannelCreateController.add(channel);
      },
      'CHANNEL_UPDATE': (data) async {
        var channel = new Channel._cached(data, client);
        if (channel is GuildChannel) {
          channel._update(data);
        }
        _onChannelUpdateController.add(channel);
      },
      'CHANNEL_DELETE': (data) async {
        var parsedId = parseInt(data['id']);
        var channel = client.state.resolveChannel(parsedId);
        if (channel == null) return;
        if (channel is GuildChannel) {
          channel.guild.channels.remove(channel.id);
          client.state.guildChannels.remove(channel.id);
        } else if (channel is DMChannel) {
          client.state.dmChannels.remove(parsedId);
        }
        _onChannelDeleteController.add(channel);
      },
      'MESSAGE_CREATE': (data) async {
        var message = new Message._create(data, client);
        message._update(data);
        _onMessageReceiveController.add(message);
      },
      'MESSAGE_UPDATE': (data) async {
        var parsedId = parseInt(data['id']);
        var parsedChannelId = parseInt(data['channel_id']);
        var channel =
        client.state.resolveChannel(parsedChannelId) as TextChannel;
        if (channel == null) return;
        if (channel.messages.containsKey(parsedId)) {
          channel.messages[parsedId]._update(data);
        }
        _onMessageUpdateController.add(channel.messages[parsedId]);
      },
      'MESSAGE_DELETE': (data) async {
        var parsedId = parseInt(data['id']);
        var parsedChannelId = parseInt(data['channel_id']);

        var channel =
        client.state.resolveChannel(parsedChannelId) as TextChannel;
        if (channel == null) return;
        var message = channel.messages[parsedId];
        channel.messages.remove(parsedId);
        _onMessageDeleteController.add(message);
      },
      'MESSAGE_DELETE_BULK': (data) async {
        var channel =
        client.state.resolveChannel(parseInt(data['channel_id']))
        as TextChannel;
        var list = <Message>[];
        await Future.forEach(data['ids'] ?? [], (jsonMessageId) {
          list.add(channel.messages[parseInt(jsonMessageId)]);
          channel.messages.remove(parseInt(jsonMessageId));
        });
        _onMessageBulkDeleteController.add(list);
      },
      'GUILD_ROLE_CREATE': (data) async {
        var guild = client.state.guilds[parseInt(data['guild_id'])];
        var role = new GuildRole._create(data['role'], guild, client);
        guild.roles.add(role);
        client.state.roles[role.id] = role;
        _onGuildRoleCreateController.add(role);
      },
      'GUILD_ROLE_UPDATE': (data) async {
        var guild = client.state.guilds[parseInt(data['guild_id'])];
        var role = guild.roles
            .where((GuildRole r) => r.id == parseInt(data['role']['id']))
            .first;
        if (role != null) {
          role._update(data['role']);
          _onGuildRoleUpdateController.add(role);
        }
      },
      'GUILD_ROLE_DELETE': (data) async {
        var guild = client.state.guilds[parseInt(data['guild_id'])];
        var role =
            guild.roles
                .where((r) => r.id == parseInt(data['role_id']))
                .first;
        guild.roles.remove(role);
        client.state.roles.remove(role);
        _onGuildRoleDeleteController.add(role);
      },
      'GUILD_MEMBER_ADD': (data) async {
        var guild = client.state.guilds[parseInt(data['guild_id'])];
        var member = new GuildMember._create(data, guild, client);
        _onGuildMemberJoinController.add(member);
      },
      'GUILD_MEMBER_UPDATE': (data) async {
        var guild = client.state.guilds[parseInt(data['guild_id'])];
        var member = guild.members[parseInt(data['user']['id'])];
        if (member != null) {
          member._update(data);
          _onGuildMemberUpdateController.add(member);
        }
      },
      'GUILD_MEMBER_REMOVE': (data) async {
        var guild = client.state.guilds[parseInt(data['guild_id'])];
        var member = guild.members[parseInt(data['user']['id'])];
        guild.members.remove(member.id);
        client.state.members.remove(member.id);
        _onGuildMemberLeaveController.add(member);
      },
      'GUILD_MEMBERS_CHUNK': (data) async {
        var guild = client.state.guilds[parseInt(data['guild_id'])];
        await Future.forEach(data['members'], (jsonMember) async {
          var member = new GuildMember._create(jsonMember, guild, client);
          client.state.members[member.id] = member;
          guild.members[member.id] = member;
        });
      }
    };
  }

  /// Handles a raw Discord event from Discord.
  Future _handle(String eventName, Map<String, dynamic> data) async {
    var handler = _eventHandlers[eventName];
    if (handler != null) {
      try {
        await handler(data);
      } on Exception catch (e) {
        // Issue #2: This currently is completely ignored and handler exceptions kill the WebSocket due to Stream async
        new Logger('Selene-Dispatch').severe(
            'A stream subscription for ${eventName} has thrown an exception: ${e
                .runtimeType} -> ${e.toString()}');
      }
    }
  }
}
