part of selene;

/// A mixin that enables message CRUD operations.
abstract class TextChannel implements SnowflakeEntity, Channel {
  /// Sends a message to this channel.
  ///
  /// Must be connected to the Real-Time Gateway.
  /// Either `content` or `embed` must be provided (or both), otherwise an [ArgumentError] will be raised.
  Future<Message> sendMessage(
      {String content, EmbedBuilder embed, bool isTTS = false}) async {
    if (content == null && embed == null) {
      throw new ArgumentError(
          'sendMessage: Content and/or embed must be provided.');
    }

    var requestParams = <String, dynamic>{'tts': isTTS};
    if (content != null) requestParams['content'] = content;
    if (embed != null) requestParams['embed'] = embed.build();

    var response = await client._rest
        .makeNewRequest('/channels/$id/messages', 'POST', requestParams);
    return new Message._create(response.body.asJson(), client);
  }

  /// Deletes a message from this channel by ID.
  ///
  /// Client must have the `MANAGE_MESSAGES` permission.
  Future deleteMessageWithId(int messageId) =>
      client._rest
          .makeNewRequest('/channels/$id/messages/$messageId', 'DELETE');

  /// Deletes a message from this channel from a message object.
  ///
  /// Client must have the `MANAGE_MESSAGES` permission.
  Future deleteMessage(Message message) => deleteMessageWithId(message.id);

  /// Edits a message by ID.
  ///
  /// Client must be the sender of the message.
  Future<Message> editMessageWithId(int messageId,
      {String content, Map<String, dynamic> embed}) async {
    var props = fillNotNullProperties({'content': content, 'embed': embed});
    if (props.length == 0)
      throw new ArgumentError(
          'editMessage: At least one of [content, embed] must be provided');
    var response = await client._rest
        .makeNewRequest('/channels/$id/messages/$messageId', 'PATCH', props);

    return new Message._create(response.body.asJson(), client);
  }

  /// Edits a message from a message object.
  ///
  /// Client must be the sender of the message.
  Future<Message> editMessage(Message message,
      {String content, Map<String, dynamic> embed}) =>
      editMessageWithId(message.id, content: content, embed: embed);

  /// Deletes a group of messages by their IDs.
  ///
  /// Messages cannot be older than two weeks.
  ///
  /// Can only be used in guild channels.
  ///
  /// Requires `MANAGE_MESSAGES `permission.
  Future deleteMessagesById(List<int> ids) async {
    if (ids == null || ids.length < 2 || ids.length > 100) {
      throw new ArgumentError(
          'deleteMessages: ids parameter must be not null and between 2-100 in length');
    }

    await client._rest.makeNewRequest(
        '/channels/$id/messages/bulk-delete', 'POST', {'messages': ids});
  }

  /// Deletes a group of messages by objects.
  ///
  /// Messages cannot be older than two weeks.
  ///
  /// Can only be used in guild channels.
  ///
  /// Requires `MANAGE_MESSAGES` permission.
  Future deleteMessages(List<Message> messages) =>
      deleteMessagesById(messages.map((msg) => msg.id));

  /// Returns the messages for a channel.
  ///
  /// Requires `VIEW_CHANNEL` if on a guild channel.
  /// Will return no messages if the user does not have the `READ_MESSAGE_HISTORY` permission.
  Future<List<Message>> getMessages(
      {int before, int around, int after, int limit: 50}) async {
    int paramCounts = 0;
    var props = <String, dynamic>{};
    if (before != null) {
      paramCounts++;
      props['before'] = before.toString();
    }
    if (around != null) {
      props['around'] = around.toString();
      paramCounts++;
    }
    if (after != null) {
      paramCounts++;
      props['after'] = after.toString();
    }
    props['limit'] = limit;
    if (paramCounts != 1)
      throw new ArgumentError(
          'getMessages: Only one of [before, around, after] can be provided (one must be provided).');

    var response = await client._rest
        .makeNewRequest('/channels/$id/messages', 'GET', props);
    var rmessages = <int, Message>{};
    await Future.forEach(response.body.asJson(), (jsonMessage) {
      var msgn = new Message._create(jsonMessage, client);
      rmessages[msgn.id] = msgn;
    });

    messages.addAll(rmessages);

    return rmessages.values;
  }

  /// Returns a specific message in a channel.
  ///
  /// Requires the `READ_MESSAGE_HISTORY` permission if in a guild channel.
  Future<Message> getMessage(int messageId) async {
    var response = await client._rest
        .makeNewRequest('/channels/$id/messages/$messageId', 'GET');
    return new Message._create(response.body.asJson(), client);
  }

  /// A message cache.
  Map<int, Message> messages = {};
}
