part of selene;

/// A guild in Discord, an isolated collection of users and channels.
///
/// Referred to as 'Servers' in the UI.
class Guild extends SnowflakeEntity {
  Guild._lazy(Map<String, dynamic> model, Client client)
      : super(model, client) {
    client.state.guilds[parseInt(model['id'])] = this;
  }

  Guild._create(Map<String, dynamic> model, Client client)
      : super(model, client) {
    _onCreate(model);
  }

  factory Guild._cached(Map<String, dynamic> model, Client client) {
    return client.state.guilds[parseInt(model['id'])];
  }

  void _onCreate(Map<String, dynamic> model) {
    isLarge = model['large'] ?? false;
    isAvailable = model['unavailable'] ?? false;
    memberCount = model['member_count'] ?? -1;

    _update(model);

    for (var jsonMember in model['members'] ?? []) {
      new GuildMember._cachedOrNew(jsonMember, this, client);
    }

    for (var jsonChannel in model['channels'] ?? []) {
      new GuildChannel._cacheOrNew(jsonChannel, this, client);
    }

    if (model['joined_at'] != null)
      dateJoined = DateTime.parse(model['joined_at']);

    client.state.guilds[id] = this;
  }

  void _update(Map<String, dynamic> model) {
    // Unavailable (received during Ready, or guild goes offline)
    if (model['unavailable'] != null) {
      isAvailable = !(model['unavailable']);
      if (isNotAvailable) {
        // Delete all local data - guild offline
        name = null;
        iconHash = null;
        splashHash = null;
        ownerId = null;
        permissionsNumber = null;
        voiceRegion = null;
        afkChannelId = null;
        afkTimeout = null;
        isEmbeddable = null;
        embeddedChannelId = null;
        verificationLevel = null;
        defaultMessageNotificationsLevel = null;
        explicitContentFilterLevel = null;
        emotes = [];
        roles = [];
        channels = {};
        members = {};
        featuresEnabled = [];
        mfaLevel = null;
        creatorApplicationId = null;
        isWidgetEnabled = null;
        widgetChannelId = null;
        systemMessageChannelId = null;
        dateJoined = null;
        isLarge = null;
        return;
      }
    }

    name = model['name'] ?? name;
    iconHash = model['icon'] ?? iconHash;
    splashHash = model['splash'] ?? splashHash;
    ownerId = model['owner_id'] ?? ownerId;
    permissionsNumber = model['permissions'] ?? permissionsNumber;
    voiceRegion = model['region'] ?? voiceRegion;
    afkChannelId = model['afk_channel_id'] ?? afkChannelId;
    afkTimeout = model['afk_timeout'] ?? afkTimeout;
    isEmbeddable = model['embed_enabled'] ?? isEmbeddable;
    embeddedChannelId = model['embed_channel_id'] ?? embeddedChannelId;
    verificationLevel = model['verification_level'] ?? verificationLevel;
    memberCount = model['member_count'] ?? memberCount;
    defaultMessageNotificationsLevel = model['default_message_notifications'] ??
        defaultMessageNotificationsLevel;
    explicitContentFilterLevel =
        model['explicit_content_filter'] ?? explicitContentFilterLevel;

    if (roles.isEmpty) {
      for (var jsonRole in model['roles'] ?? []) {
        var role = new GuildRole._create(jsonRole, this, client);
        roles.add(role);
        client.state.roles[role.id] = role;
      }
    }

    for (var jsonEmote in model['emojis'] ?? []) {
      var emote = new GuildEmote(jsonEmote, this, client);
      emotes.add(emote);
    }

    for (var feature in model['features'] ?? []) {
      if (!featuresEnabled.contains(feature.toString()))
        featuresEnabled.add(feature.toString());
    }

    mfaLevel = model['mfa_level'] ?? mfaLevel;
    creatorApplicationId = model['application_id'] ?? creatorApplicationId;
    isWidgetEnabled = model['widget_enabled'] ?? isWidgetEnabled;
    widgetChannelId = model['widget_channel_id'] ?? widgetChannelId;
    systemMessageChannelId =
        model['system_channel_id'] ?? systemMessageChannelId;
  }

  /// The name of this guild (2-100 characters).
  String name;

  /// The hash of this guild's icon.
  ///
  /// Null if this server does not have an icon.
  String iconHash;

  /// The hash of this guild's splash image.
  ///
  /// Null if this server has no splash.
  String splashHash;

  /// The ID of the owner of this guild.
  String ownerId;

  /// The total permissions for the requester.
  ///
  /// Does not include channel overrides.
  int permissionsNumber;

  /// The voice region of this guild.
  String voiceRegion;

  /// The ID of the AFK channel for this guild.
  ///
  /// Null if it doesn't exist.
  String afkChannelId;

  /// The AFK timeout for this guild in seconds.
  int afkTimeout;

  /// Whether or not this guild has an AFK channel.
  bool get hasAfkChannel => afkChannelId != null;

  /// Whether or not this guild does not have an AFK channel.
  bool get doesNotHaveAfkChannel => !hasAfkChannel;

  /// If this guild can be embedded (i.e. in a widget)
  bool isEmbeddable;

  /// If this guild can not be embedded (i.e. in a widget)
  bool get isNotEmbeddable => !isEmbeddable;

  /// The ID of the embedded channel.
  ///
  /// Null if this is not embedded.
  int embeddedChannelId;

  /// The verification level required for this guild.
  ///
  /// See https://discordapp.com/developers/docs/resources/guild#guild-object-verification-level
  int verificationLevel;

  /// The default message notifications option set for this guild.
  ///
  /// See https://discordapp.com/developers/docs/resources/guild#guild-object-default-message-notification-level
  int defaultMessageNotificationsLevel;

  /// The level that the explicit content filter operates at.
  ///
  /// See https://discordapp.com/developers/docs/resources/guild#guild-object-explicit-content-filter-level
  int explicitContentFilterLevel;

  /// A list of roles in this guild.
  List<GuildRole> roles = [];

  /// A list of all custom emotes in this guild.
  List<Emote> emotes = [];

  /// A list of features enabled for this guild.
  List<String> featuresEnabled = [];

  /// The required MFA level for this guild.
  ///
  /// See https://discordapp.com/developers/docs/resources/guild#guild-object-mfa-level
  int mfaLevel;

  /// The application ID of the guild creator, if it is bot-created.
  String creatorApplicationId;

  /// Whether this guild is bot-created.
  bool get isCreatedByBot => creatorApplicationId != null;

  /// Whether this guild is not bot-created.
  bool get isNotCreatedByBot => !isCreatedByBot;

  /// Whether this guild's widget is enabled.
  bool isWidgetEnabled;

  /// Whether this guild's widget is not enabled.
  bool get isNotWidgetEnabled => !isWidgetEnabled;

  /// If the widget is enabled, the channel ID for it.
  String widgetChannelId;

  /// The ID of the channel where system messages are sent.
  String systemMessageChannelId;

  /// When the requester joined this guild.
  DateTime dateJoined;

  /// Whether this guild is considered 'large'.
  bool isLarge;

  /// Whether this guild is not considered 'large'.
  bool get isNotLarge => !isLarge;

  /// Whether or not this guild is available.
  bool isAvailable;

  /// Whether or not this guild is not available (unavailable).
  bool get isNotAvailable => !isAvailable;

  /// The total count of members in this guild.
  int memberCount;

  /// The channels in this guild.
  Map<int, GuildChannel> channels = {};

  /// All the members of this guild that have been downloaded or received.
  Map<int, GuildMember> members = {};

  // TODO: Voice States
  // TODO: Presences

  /// Deletes this guild permanently.
  ///
  /// The current user must be the owner.
  Future delete([RequestOptions options]) async {
    if (ownerId != client.currentUser.id.toString())
      throw new UnsupportedError('Cannot delete guild if not owner');

    await client._rest.makeNewRequest('/guilds/$id', 'DELETE', null, options);
  }

  /// Modifies the properties of this guild.
  ///
  /// The current user must have the 'MANAGE_GUILD' permission.
  Future modify({String name,
    VoiceRegion region,
    int verificationLevel,
    int defaultMessageNotifications,
    int explicitContentFilter,
    GuildChannel afkChannel,
    int afkTimeout,
    String icon_imageBase64_128x128_jpg,
    GuildMember owner,
    String splash_imageBase64_128x128_jpg,
    GuildChannel systemChannel,
    RequestOptions options}) async {
    var props = fillNotNullProperties({
      'name': name,
      'verification_level': verificationLevel,
      'default_message_notifications': defaultMessageNotifications,
      'explicit_content_filter': explicitContentFilter,
      'afk_timeout': afkTimeout,
      'icon': icon_imageBase64_128x128_jpg,
      'splash': splash_imageBase64_128x128_jpg
    });
    if (region != null) props['region'] = region.id;
    if (afkChannel != null) props['afk_channel_id'] = afkChannel.id;
    if (owner != null) {
      if (ownerId != client.currentUser.id.toString())
        throw new UnsupportedError(
            'Cannot switch guild ownership if not owner');
      props['owner_id'] = owner.id;
    }
    if (systemChannel != null) props['system_channel_id'] = systemChannel.id;
    var response = await client._rest.makeNewRequest(
        '/guilds/$id', 'PATCH', props, options);
    return new Guild._create(response.body.asJson(), client);
  }

  String toString() => "${name} (ID ${id})";
}
